import rbush from 'rbush';
import Coordinate from 'jsts/org/locationtech/jts/geom/Coordinate';
import GeometryFactory from 'jsts/org/locationtech/jts/geom/GeometryFactory';
import IsSimpleOp from 'jsts/org/locationtech/jts/operation/IsSimpleOp';
import RelateOp from 'jsts/org/locationtech/jts/operation/relate/RelateOp';
import DistanceOp from 'jsts/org/locationtech/jts/operation/distance/DistanceOp';
import Polygonizer from 'jsts/org/locationtech/jts/operation/polygonize/Polygonizer';
import BoundaryNodeRule from 'jsts/org/locationtech/jts/algorithm/BoundaryNodeRule';
import LengthIndexedLine from 'jsts/org/locationtech/jts/linearref/LengthIndexedLine';

const factory = new GeometryFactory();
const isSimpleOp = new IsSimpleOp();

function toLineString (coordinates) {
  return factory.createLineString(coordinates.map(c => new Coordinate(c[0], c[1])))
}

function toCoordinate (c) {
  return new Coordinate(c[0], c[1])
}

function toPoint (c) {
  return factory.createPoint(new Coordinate(c[0], c[1]))
}

function polyToCoordss (poly) {
  const cs = lineStringToCoords(poly.getExteriorRing());
  return [cs]
}

function lineStringToCoords (ls) {
  return ls.getCoordinates().map(c => [c.x, c.y])
}

function isSimple (coordinates) {
  const lineString = toLineString(coordinates);
  return isSimpleOp.isSimpleLinearGeometry(lineString)
}

function relate (cs1, cs2) {
  const ls1 = toLineString(cs1);
  const ls2 = toLineString(cs2);
  return RelateOp.relate(ls1, ls2, BoundaryNodeRule.ENDPOINT_BOUNDARY_RULE)
}

function intersects (cs1, cs2) {
  const ls1 = toLineString(cs1);
  const ls2 = toLineString(cs2);
  return RelateOp.intersects(ls1, ls2)
}

function equals (c1, c2) {
  return c1[0] === c2[0] && c1[1] === c2[1]
}

function signedArea (shell) {
  if (shell.length < 3) {
    return 0
  }
  let sum = 0;
  let x;
  let y1;
  let y2;
  let p1 = shell[0];
  let p2 = shell[1];
  const x0 = p1[0];
  for (let i = 2; i < shell.length; i++) {
    let p3 = shell[i];
    x = p2[0] - x0;
    y1 = p3[1];
    y2 = p1[1];
    sum += x * (y2 - y1);
    p1 = p2;
    p2 = p3;
  }
  return sum / 2
}

function azimuth (a, b) {
  let d;

  if (a[0] === b[0]) {
    if (a[1] < b[1]) d = 0.0;
    else if (a[1] > b[1]) d = Math.PI;
    else throw new Error('same coordinate')
    return d
  }

  if (a[1] === b[1]) {
    if (a[0] < b[0]) d = Math.PI / 2;
    else if (a[0] > b[0]) d = Math.PI + (Math.PI / 2);
    else throw new Error('same coordinate')
    return d
  }

  if (a[0] < b[0]) {
    if (a[1] < b[1]) {
      d = Math.atan(Math.abs(a[0] - b[0]) / Math.abs(a[1] - b[1]));
    } else {
      d = Math.atan(Math.abs(a[1] - b[1]) / Math.abs(a[0] - b[0])) + (Math.PI / 2);
    }
  } else {
    if (a[1] > b[1]) {
      d = Math.atan(Math.abs(a[0] - b[0]) / Math.abs(a[1] - b[1])) + Math.PI;
    } else {
      d = Math.atan(Math.abs(a[1] - b[1]) / Math.abs(a[0] - b[0])) + (Math.PI + (Math.PI / 2));
    }
  }

  return d
}

function polygonize (css) {
  const lss = css.map(toLineString);
  const polygonizer = new Polygonizer();
  lss.forEach(ls => polygonizer.add(ls));
  const polys = polygonizer.getPolygons();
  if (polys.size() === 0) {
    throw new Error('Could not polygonize edges - invalid topology?')
  }
  return polyToCoordss(polys.get(0))
}

function split (coordinates, coordinate) {
  const ls = toLineString(coordinates);
  const c = toCoordinate(coordinate);
  const lil = new LengthIndexedLine(ls);
  const splitIndex = lil.project(c);
  const ls1 = lil.extractLine(lil.getStartIndex(), splitIndex);
  const ls2 = lil.extractLine(splitIndex, lil.getEndIndex());
  const cs1 = lineStringToCoords(ls1);
  const cs2 = lineStringToCoords(ls2);
  return [cs1, cs2]
}

function isLeft (c0, c1, c2) {
  return ((c1[0] - c0[0]) * (c2[1] - c0[1])) - ((c2[0] - c0[0]) * (c1[1] - c0[1]))
}

function distance (c, cs) {
  const point = toPoint(c);
  const lineString = toLineString(cs);
  return DistanceOp.distance(point, lineString)
}

function pointInPoly (c, shell) {
  return calcWindingNumber(c, shell) !== 0
}

function calcWindingNumber (c, shell) {
  let wn = 0;
  for (let i = 0; i < shell.length - 1; i++) {
    const va = shell[i];
    const vb = shell[i + 1];
    if (va[1] <= c[1]) {
      if (vb[1] > c[1]) {
        const l = isLeft(va, vb, c);
        if (l > 0) wn++;
        else if (l === 0) return 0
      }
    } else {
      if (vb[1] <= c[1]) {
        const l = isLeft(va, vb, c);
        if (l < 0) wn--;
        else if (l === 0) return 0
      }
    }
  }
  return wn
}

class SpatialError extends Error {}

/** @module */

/**
 * Edge definition
 *
 * @typedef {Object} Edge
 * @property {number} id Edge ID
 * @property {module:node~Node} start
 * @property {module:node~Node} end
 * @property {module:coordinate~Coordinate[]} coordinates Coordinates
 * @property {module:edge~Edge} nextLeft
 * @property {module:edge~Edge} nextRight
 * @property {module:face~Face} leftFace
 * @property {module:face~Face} rightFace
 * @property {number} minX Minimum X of bounds
 * @property {number} maxY Maximum Y of bounds
 * @property {number} minX Minimum X of bounds
 * @property {number} maxY Maximum Y of bounds
 */

console.debug = console.log;
console.debug = function () {};

/**
 * @param {module:edge~Edge} e
 * @param {boolean} d
 * @return {number}
 * @private
 */
function sid (e, d) {
  return d ? e.id : -e.id
}

/**
 * @param {module:edge~Edge} e
 * @return {string}
 * @private
 */


function getEdgeByPoint (topo, c, tol) {
  const result = topo.edgesTree.search({
    minX: c[0] - tol,
    minY: c[1] - tol,
    maxX: c[0] + tol,
    maxY: c[1] + tol
  });

  const candidates = result.filter(e => distance(c, e.coordinates) < tol);

  // TODO: throw exception on more than one candidate?

  return candidates
}

function getEdgesByLine (topo, cs) {
  const xs = cs.map(c => c[0]);
  const ys = cs.map(c => c[1]);

  const bounds = {
    minX: Math.min(...xs),
    minY: Math.min(...ys),
    maxX: Math.max(...xs),
    maxY: Math.max(...ys)
  };

  const edges = topo.edgesTree.search(bounds)
    .filter(e => intersects(cs, e.coordinates));

  return edges
}

/**
 * Adds an isolated edge defined by geometry alinestring to a topology connecting two existing isolated nodes anode and anothernode and returns the edge id of the new edge.
 *
 * @param {module:topo~Topo} topo
 * @param {module:node~Node} start
 * @param {module:node~Node} end
 * @param {module:coordinate~Coordinate[]} coordinates
 * @return {module:edge~Edge}
 */
function addIsoEdge (topo, start, end, coordinates) {
  const xs = coordinates.map(c => c[0]);
  const ys = coordinates.map(c => c[1]);

  const edge = {
    id: topo.edgesSeq,
    start,
    end,
    coordinates,
    nextLeft: { id: 0 },
    nextRight: { id: 0 },
    leftFace: { id: -1 },
    rightFace: { id: -1 },
    minX: Math.min(...xs),
    minY: Math.min(...ys),
    maxX: Math.max(...xs),
    maxY: Math.max(...ys)
  };

  if (start === end) {
    throw new SpatialError('start and end node cannot be the same as it would not construct an isolated edge')
  }

  if (!start.face || !end.face) {
    throw new SpatialError('not isolated node')
  }

  if (start.face !== end.face) {
    throw new SpatialError('nodes in different faces')
  }

  if (!equals(start.coordinate, coordinates[0])) {
    throw new SpatialError('start node not geometry start point')
  }

  if (!equals(end.coordinate, coordinates[coordinates.length - 1])) {
    throw new SpatialError('end node not geometry end point')
  }

  if (!isSimple(coordinates)) {
    throw new SpatialError('curve not simple')
  }

  checkEdgeCrossing(topo, start, end, edge);

  edge.leftFace = end.face;
  edge.nextLeft = edge;
  edge.nextLeftDir = false;
  edge.nextRight = edge;
  edge.nextRightDir = true;

  delete start.face;
  delete end.face;

  insertEdge(topo, edge);
  trigger(topo, 'addedge', edge);

  return edge
}

function checkEdgeCrossing (topo, start, end, edge) {
  const check = (e1, e2) => {
    if (e1 === e2) {
      return
    }
    const im = relate(e1.coordinates, e2.coordinates);
    if (im.matches('1FFF*FFF2')) {
      throw new SpatialError('coincident edge ' + e1.id)
    }
    if (im.matches('1********')) {
      throw new SpatialError('geometry intersects edge ' + e1.id)
    }
    if (im.matches('T********')) {
      throw new SpatialError('geometry crosses edge ' + e1.id)
    }
  };
  topo.edgesTree.search(edge).forEach(e => check(e, edge));
}

function getEdgeByNode (topo, node) {
  if (node.length) {
    return topo.edges.filter(e => node.some(n => n === e.start) || node.some(n => n === e.end))
  } else {
    return topo.edges.filter(e => e.start === node || e.end === node)
  }
}

function findAdjacentEdges (topo, node, data, other, edge) {
  data.nextCW = data.nextCCW = { id: 0 };
  data.cwFace = data.ccwFace = { id: -1 };

  let minaz, maxaz, azdif;

  if (other) {
    azdif = other.az - data.az;
    if (azdif < 0) azdif += 2 * Math.PI;
    minaz = maxaz = azdif;
    console.debug(`Other edge end has cwFace=${other.cwFace.id} and ccwFace=${other.ccwFace.id}`);
  } else {
    minaz = maxaz = -1;
  }

  console.debug(`Looking for edges incident to node ${node.id} and adjacent to azimuth ${data.az}`);

  const edges = getEdgeByNode(topo, node);

  console.debug(`getEdgeByNode returned ${edges.length} edges, minaz=${minaz}, maxaz=${maxaz}`);

  edges.forEach(e => {
    if (e === edge) {
      return
    }

    if (e.coordinates.length < 2) {
      throw new Error(`corrupted topo: edge ${e.id} does not have two distinct points`)
    }

    if (e.start === node) {
      const p1 = e.coordinates[0];
      const p2 = e.coordinates[1];
      console.debug(`edge ${e.id} starts on node ${node.id}, edgeend is ${p1[0]},${p1[1]}-${p2[0]},${p2[1]}`);
      const az = azimuth(p1, p2);
      azdif = az - data.az;
      console.debug(`azimuth of edge ${e.id}: ${az} (diff: ${azdif})`);
      if (azdif < 0) azdif += 2 * Math.PI;
      if (minaz === -1) {
        minaz = maxaz = azdif;
        data.nextCW = data.nextCCW = e;
        data.nextCWDir = data.nextCCWDir = true;
        data.cwFace = e.leftFace;
        data.ccwFace = e.rightFace;
        console.debug(`new nextCW and nextCCW edge is ${e.id}, outgoing, with face_left ${e.leftFace.id} and face_right ${e.rightFace.id} (face_right is new ccwFace, face_left is new cwFace)`);
      } else {
        if (azdif < minaz) {
          data.nextCW = e;
          data.nextCWDir = true;
          data.cwFace = e.leftFace;
          console.debug(`new nextCW edge is ${e.id}, outgoing, with face_left ${e.leftFace.id} and face_right ${e.rightFace.id} (previous had minaz=${minaz}, face_left is new cwFace)`);
          minaz = azdif;
        } else if (azdif > maxaz) {
          data.nextCCW = e;
          data.nextCCWDir = true;
          data.ccwFace = e.rightFace;
          console.debug(`new nextCCW edge is ${e.id}, outgoing, with face_left ${e.leftFace.id} and face_right ${e.rightFace.id} (previous had maxaz=${maxaz}, face_right is new ccwFace)`);
          maxaz = azdif;
        }
      }
    }

    if (e.end === node) {
      const p1 = e.coordinates[e.coordinates.length - 1];
      const p2 = e.coordinates[e.coordinates.length - 2];
      console.debug(`edge ${e.id} ends on node ${node.id}, edgeend is ${p1[0]},${p1[1]}-${p2[0]},${p2[1]}`);
      const az = azimuth(p1, p2);
      azdif = az - data.az;
      console.debug(`azimuth of edge ${e.id}: ${az} (diff: ${azdif})`);
      if (azdif < 0) azdif += 2 * Math.PI;
      if (minaz === -1) {
        minaz = maxaz = azdif;
        data.nextCW = data.nextCCW = e;
        data.nextCWDir = data.nextCCWDir = false;
        data.cwFace = e.rightFace;
        data.ccwFace = e.leftFace;
        console.debug(`new nextCW and nextCCW edge is ${e.id}, incoming, with face_left ${e.leftFace.id} and face_right ${e.rightFace.id} (face_right is new cwFace, face_left is new ccwFace)`);
      } else {
        if (azdif < minaz) {
          data.nextCW = e;
          data.nextCWDir = false;
          data.cwFace = e.rightFace;
          console.debug(`new nextCW edge is ${e.id}, incoming, with face_left ${e.leftFace.id} and face_right ${e.rightFace.id} (previous had minaz=${minaz}, face_right is new cwFace)`);
          minaz = azdif;
        } else if (azdif > maxaz) {
          data.nextCCW = e;
          data.nextCCWDir = false;
          data.ccwFace = e.leftFace;
          console.debug(`new nextCCW edge is ${e.id}, outgoing, from start point, with face_left ${e.leftFace.id} and face_right ${e.rightFace.id} (previous had maxaz=${maxaz}, face_left is new ccwFace)`);
          maxaz = azdif;
        }
      }
    }
  });

  console.debug(`edges adjacent to azimuth ${data.az} (incident to node ${node.id}): CW:${sid(data.nextCW, data.nextCWDir)}(${minaz}) CCW:${sid(data.nextCCW, data.nextCCWDir)}(${maxaz})`);

  if (!edge && edges.length > 0 && data.cwFace !== data.ccwFace) {
    if (data.cwFace.id !== -1 && data.ccwFace.id !== -1) {
      throw new Error(`Corrupted topo: adjacent edges ${sid(data.nextCW, data.nextCWDir)} and ${sid(data.nextCCW, data.nextCCWDir)} bind different face (${data.cwFace.id} and ${data.ccwFace.id})`)
    }
  }

  return edges
}

function addEdge (topo, start, end, coordinates, modFace) {
  console.debug('addEdge called');

  if (!isSimple(coordinates)) {
    throw new SpatialError('curve not simple')
  }

  const xs = coordinates.map(c => c[0]);
  const ys = coordinates.map(c => c[1]);

  const edge = {
    id: topo.edgesSeq,
    start,
    end,
    coordinates,
    nextLeft: { id: 0 },
    nextRight: { id: 0 },
    leftFace: { id: -1 },
    rightFace: { id: -1 },
    minX: Math.min(...xs),
    minY: Math.min(...ys),
    maxX: Math.max(...xs),
    maxY: Math.max(...ys)
  };

  // TODO: remove repeated points
  // TODO: check that we haave at least two points left

  const span = {
    cwFace: { id: -1 },
    ccwFace: { id: -1 },
    az: azimuth(coordinates[0], coordinates[1])
  };

  const epan = {
    cwFace: { id: -1 },
    ccwFace: { id: -1 },
    az: azimuth(coordinates[coordinates.length - 1], coordinates[coordinates.length - 2])
  };

  const nodes = start !== end ? [start, end] : [start];

  nodes.forEach(node => {
    if (node.face) {
      if (edge.leftFace.id === -1) {
        edge.leftFace = edge.rightFace = node.face;
      } else if (edge.leftFace !== node.face) {
        throw new SpatialError(`geometry crosses an edge (endnodes in faces ${edge.leftFace.id} and ${node.face.id})`)
      }
    }
  });

  if (!equals(start.coordinate, coordinates[0])) {
    throw new SpatialError('start node not geometry start point')
  }

  if (!equals(end.coordinate, coordinates[coordinates.length - 1])) {
    throw new SpatialError('end node not geometry end point')
  }

  checkEdgeCrossing(topo, start, end, edge);

  const isClosed = start === end;
  const foundStart = findAdjacentEdges(topo, start, span, isClosed ? epan : undefined);

  let prevLeft;
  let prevLeftDir;

  if (foundStart.length > 0) {
    span.wasIsolated = false;
    if (span.nextCW.id) {
      edge.nextRight = span.nextCW;
      edge.nextRightDir = span.nextCWDir;
    } else {
      edge.nextRight = edge;
      edge.nextRightDir = false;
    }
    if (span.nextCCW.id) {
      prevLeft = span.nextCCW;
      prevLeftDir = !span.nextCCWDir;
    } else {
      prevLeft = edge;
      prevLeftDir = true;
    }
    console.debug(`New edge ${edge.id} is connected on start node, next_right is ${sid(edge.nextRight, edge.nextRightDir)}, prev_left is ${sid(prevLeft, prevLeftDir)}`);
    if (edge.rightFace.id === -1) {
      edge.rightFace = span.cwFace;
    }
    if (edge.leftFace.id === -1) {
      edge.leftFace = span.ccwFace;
    }
  } else {
    span.wasIsolated = true;
    edge.nextRight = edge;
    edge.nextRightDir = !isClosed;
    prevLeft = edge;
    prevLeftDir = isClosed;
    console.debug(`New edge ${edge.id} is isolated on start node, next_right is ${sid(edge.nextRight, edge.nextRightDir)}, prev_left is ${sid(prevLeft, prevLeftDir)}`);
  }

  const foundEnd = findAdjacentEdges(topo, end, epan, isClosed ? span : undefined);

  let prevRight;
  let prevRightDir;

  if (foundEnd.length > 0) {
    epan.wasIsolated = false;
    if (epan.nextCW.id) {
      edge.nextLeft = epan.nextCW;
      edge.nextLeftDir = epan.nextCWDir;
    } else {
      edge.nextLeft = edge;
      edge.nextLeftDir = true;
    }
    if (epan.nextCCW.id) {
      prevRight = epan.nextCCW;
      prevRightDir = !epan.nextCCWDir;
    } else {
      prevRight = edge;
      prevRightDir = false;
    }
    console.debug(`New edge ${edge.id} is connected on end node, next_left is ${sid(edge.nextLeft, edge.nextLeftDir)}, prev_right is ${sid(prevRight, prevRightDir)}`);
    if (edge.rightFace.id === -1) {
      edge.rightFace = span.ccwFace;
    } else if (edge.rightFace !== epan.ccwFace) {
      throw new Error(`Side-location conflict: new edge starts in face ${edge.rightFace.id} and ends in face ${epan.ccwFace.id}`)
    }
    if (edge.leftFace.id === -1) {
      edge.leftFace = span.cwFace;
    } else if (edge.leftFace !== epan.cwFace) {
      throw new Error(`Side-location conflict: new edge starts in face ${edge.leftFace.id} and ends in face ${epan.cwFace.id}`)
    }
  } else {
    epan.wasIsolated = true;
    edge.nextLeft = edge;
    edge.nextLeftDir = isClosed;
    prevRight = edge;
    prevRightDir = !isClosed;
    console.debug(`New edge ${edge.id} is isolated on end node, next_left is ${sid(edge.nextLeft, edge.nextLeftDir)}, prev_right is ${sid(prevRight, prevRightDir)}`);
  }

  if (edge.leftFace !== edge.rightFace) {
    throw new Error(`Left(${edge.leftFace.id})/right(${edge.rightFace.id}) faces mismatch: invalid topology ?`)
  } else if (edge.leftFace.id === -1) {
    throw new Error('Could not derive edge face from linked primitives: invalid topo ?')
  }

  insertEdge(topo, edge);

  if (prevLeft !== edge) {
    if (prevLeftDir) {
      prevLeft.nextLeft = edge;
      prevLeft.nextLeftDir = true;
    } else {
      prevLeft.nextRight = edge;
      prevLeft.nextRightDir = true;
    }
  }

  if (prevRight !== edge) {
    if (prevRightDir) {
      prevRight.nextLeft = edge;
      prevRight.nextLeftDir = false;
    } else {
      prevRight.nextRight = edge;
      prevRight.nextRightDir = false;
    }
  }

  if (span.wasIsolated) {
    delete start.face;
  }
  if (epan.wasIsolated) {
    delete end.face;
  }

  if (!isClosed && (epan.wasIsolated || span.wasIsolated)) {
    trigger(topo, 'addedge', edge);
    return edge
  }

  const oldFace = edge.leftFace;

  let newface1;

  if (!modFace) {
    newface1 = addFaceSplit(topo, edge, false, edge.leftFace, false);
    if (newface1 === 0) {
      console.debug('New edge does not split any face');
      trigger(topo, 'addedge', edge);
      return edge
    }
  }

  let newface = addFaceSplit(topo, edge, true, edge.leftFace, false);

  if (modFace) {
    if (newface === 0) {
      console.debug('New edge does not split any face');
      trigger(topo, 'addedge', edge);
      return edge
    }

    if (newface < 0) {
      newface = addFaceSplit(topo, edge, false, edge.leftFace, false);
      if (newface < 0) {
        trigger(topo, 'addedge', edge);
        return edge
      }
    } else {
      addFaceSplit(topo, edge, false, edge.leftFace, true);
    }
  }

  trigger(topo, 'addedge', edge);

  if (oldFace !== topo.universe && !modFace) {
    deleteFace(topo, oldFace);
    trigger(topo, 'removeface', oldFace);
  }

  return edge
}

/**
 * Add a new edge and, if in doing so it splits a face, delete the original face and replace it with two new faces.
 *
 * @param {module:topo~Topo} topo
 * @param {module:node~Node} start
 * @param {module:node~Node} end
 * @param {module:coordinate~Coordinate[]} coordinates
 * @return {module:edge~Edge}
 */
function addEdgeNewFaces (topo, start, end, coordinates) {
  return addEdge(topo, start, end, coordinates, false)
}

/**
 * Add a new edge and, if in doing so it splits a face, modify the original face and add a new face.
 *
 * @param {module:topo~Topo} topo
 * @param {module:node~Node} start
 * @param {module:node~Node} end
 * @param {module:coordinate~Coordinate[]} coordinates
 * @return {module:edge~Edge}
 */
function addEdgeModFace (topo, start, end, coordinates) {
  return addEdge(topo, start, end, coordinates, true)
}

function remEdge (topo, edge, modFace) {
  console.debug('Updating next_{right,left}_face of ring edges...');

  const { universe, edges, nodes, facesTree } = topo;

  const oldLeftFace = edge.leftFace;
  const oldRightFace = edge.rightFace;

  const edgeNodes = [];
  edgeNodes.push(edge.start);
  if (edge.end !== edge.start) {
    edgeNodes.push(edge.end);
  }
  const updEdge = getEdgeByNode(topo, edgeNodes);
  const updEdgeLeft = [];
  const updEdgeRight = [];
  let fnodeEdges = 0;
  let lnodeEdges = 0;

  updEdge.forEach(e => {
    if (e === edge) return
    if (e.start === edge.start || e.end === edge.start) fnodeEdges++;
    if (e.start === edge.end || e.end === edge.end) lnodeEdges++;
    if (e.nextLeft === edge && !e.nextLeftDir) {
      updEdgeLeft.push({
        edge: e,
        nextLeft: edge.nextLeft !== edge || !edge.nextLeftDir ? edge.nextLeft : edge.nextRight,
        nextLeftDir: edge.nextLeft !== edge || !edge.nextLeftDir ? edge.nextLeftDir : edge.nextRightDir
      });
    } else if (e.nextLeft === edge && e.nextLeftDir) {
      updEdgeLeft.push({
        edge: e,
        nextLeft: edge.nextRight !== edge || edge.nextRightDir ? edge.nextRight : edge.nextLeft,
        nextLeftDir: edge.nextRight !== edge || edge.nextRightDir ? edge.nextRightDir : edge.nextLeftDir
      });
    }

    if (e.nextRight === edge && !e.nextRightDir) {
      updEdgeRight.push({
        edge: e,
        nextRight: edge.nextLeft !== edge || !edge.nextLeftDir ? edge.nextLeft : edge.nextRight,
        nextRightDir: edge.nextLeft !== edge || !edge.nextLeftDir ? edge.nextLefttDir : edge.nextRightDir
      });
    } else if (e.nextRight === edge && e.nextRightDir) {
      updEdgeRight.push({
        edge: e,
        nextRight: edge.nextRight !== edge || edge.nextRightDir ? edge.nextRight : edge.nextLeft,
        nextRightDir: edge.nextRight !== edge || edge.nextRightDir ? edge.nextRightDir : edge.nextLeftDir
      });
    }
  });

  updEdgeLeft.forEach(ue => {
    ue.edge.nextLeft = ue.nextLeft;
    ue.edge.nextLeftDir = ue.nextLeftDir;
  });

  updEdgeRight.forEach(ue => {
    ue.edge.nextRight = ue.nextRight;
    ue.edge.nextRightDir = ue.nextRightDir;
  });

  let floodface;
  let newface = { id: -1 };

  if (oldLeftFace === oldRightFace) {
    floodface = oldRightFace;
  } else {
    if (oldLeftFace === universe || oldRightFace === universe) {
      floodface = universe;
      console.debug('floodface is universe');
    } else {
      floodface = oldRightFace;
      console.debug('floodface is ' + floodface.id);
      // TODO: merge bboxes
      if (modFace) {
        newface = floodface;
      } else {
        insertFace(topo, newface);
        floodface = newface;
      }
    }

    if (oldLeftFace !== floodface) {
      console.debug(`Updating edges leftFace to ${floodface.id} where it was ${oldLeftFace.id}`);
      edges
        .filter(e => e.leftFace === oldLeftFace)
        .forEach(e => (e.leftFace = floodface));
      edges
        .filter(e => e.rightFace === oldLeftFace)
        .forEach(e => (e.rightFace = floodface));
      nodes
        .filter(n => n.face === oldLeftFace)
        .forEach(n => (n.face = floodface));
    }

    if (oldRightFace !== floodface) {
      console.debug(`Updating edges rightFace to ${floodface.id} where it was ${oldRightFace.id}`);
      edges
        .filter(e => e.leftFace === oldRightFace)
        .forEach(e => (e.leftFace = floodface));
      edges
        .filter(e => e.rightFace === oldRightFace)
        .forEach(e => (e.rightFace = floodface));
      nodes
        .filter(n => n.face === oldRightFace)
        .forEach(n => (n.face = floodface));
    }
  }

  deleteEdge(topo, edge);

  if (!fnodeEdges) edge.start.face = floodface;

  if (edge.end !== edge.start && !lnodeEdges) edge.end.face = floodface;

  const deletedFaces = [];
  if (oldLeftFace !== oldRightFace) {
    if (oldRightFace !== floodface) deletedFaces.push(oldRightFace);
    if (oldLeftFace !== floodface) deletedFaces.push(oldLeftFace);
  }

  newface = modFace ? floodface : newface;

  deletedFaces.forEach(f => { deleteFace(topo, f); trigger(topo, 'removeface', f); });

  trigger(topo, 'removeedge', edge);
  if (newface.id !== -1) {
    facesTree.remove(newface);
    updateFaceTree(topo, newface);
  }
  trigger(topo, 'addface', newface);

  return modFace ? floodface : newface
}

function remEdgeNewFace (topo, edge) {
  return remEdge(topo, edge, false)
}

function remEdgeModFace (topo, edge) {
  return remEdge(topo, edge, true)
}

function healEdges (topo, e1, e2, modEdge) {
  return undefined
}

function modEdgeHeal (topo, e1, e2) {
  return healEdges(topo, e1, e2, true)
}

function newEdgeHeal (topo, e1, e2) {
  return healEdges(topo, e1, e2, false)
}

function modEdgeSplit (topo, edge, coordinate) {
  const { edges, edgesTree } = topo;

  const parts = split(edge.coordinates, coordinate);

  const splitCoordinate = parts[0][parts[0].length - 1];

  const node = {
    coordinate: splitCoordinate
  };

  insertNode(topo, node);

  const newedge1 = {
    start: node,
    end: edge.end,
    leftFace: edge.leftFace,
    rightFace: edge.rightFace
  };
  newedge1.nextLeft = edge.nextLeft === edge && !edge.nextLeftDir ? newedge1 : edge.nextLeft;
  newedge1.nextLeftDir = edge.nextLeft === edge && !edge.nextLeftDir ? false : edge.nextLeftDir;
  newedge1.nextRight = edge;
  newedge1.nextRightDir = false;
  newedge1.coordinates = parts[1];

  insertEdge(topo, newedge1);

  const oldEnd = edge.end;

  edge.coordinates = parts[0];
  edge.nextLeft = newedge1;
  edge.nextLeftDir = true;
  edge.end = node;
  edgesTree.remove(edge);
  const xs = edge.coordinates.map(c => c[0]);
  const ys = edge.coordinates.map(c => c[1]);
  edge.minX = Math.min(...xs);
  edge.minY = Math.min(...ys);
  edge.maxX = Math.max(...xs);
  edge.maxY = Math.max(...ys);
  edgesTree.insert(edge);

  edges
    .filter(e => e.nextRight === edge && !e.nextRightDir && e.start === oldEnd && e !== newedge1)
    .forEach(e => { e.nextRight = newedge1; e.nextRightDir = false; });

  edges
    .filter(e => e.nextLeft === edge && !e.nextLeftDir && e.end === oldEnd && e !== newedge1)
    .forEach(e => { e.nextLeft = newedge1; e.nextLeftDir = false; });

  trigger(topo, 'addnode', node);
  trigger(topo, 'addedge', newedge1);
  trigger(topo, 'modedge', edge);

  return node
}

function newEdgesSplit (topo, edge, coordinate, skipISOChecks) {
  return undefined
}

/** @module */

/**
 * Face definition
 *
 * @typedef {Object} Face
 * @property {number} id Face ID
 */

function getNodeByFace (topo, face) {
  // TODO: only within face mbr
  return topo.nodes.filter(n => n.face === face)
}

function getFaceByPoint (topo, c, tol) {
  const result = topo.facesTree.search({
    minX: c[0] - tol,
    minY: c[1] - tol,
    maxX: c[0] + tol,
    maxY: c[1] + tol
  });

  const fs = result
    .map(f => ({
      f,
      s: getFaceGeometry(topo, f)[0]
    }));

  const candidates = fs
    .filter(fs => pointInPoly(c, fs.s))
    .map(fs => fs.f);

  // TODO: throw exception on more than one candidate?

  return candidates
}

/**
 * Returns the ordered set of signed edge identifiers met by walking on an a given edge side.
 */
function getRingEdges (topo, edge, dir, limit, foundEdges) {
  foundEdges = foundEdges || [];
  foundEdges.push({ edge, dir });

  const nextEdge = dir ? edge.nextLeft : edge.nextRight;
  const nextDir = dir ? edge.nextLeftDir : edge.nextRightDir;

  if (!foundEdges.some(fe => fe.edge === nextEdge && fe.dir === nextDir)) {
    return getRingEdges(topo, nextEdge, nextDir, 0, foundEdges)
  }

  return foundEdges
}

function getEdgeByFace (topo, face, mbr) {
  return topo.edges.filter(e => e.leftFace === face || e.rightFace === face)
  // TODO: include within mbr
}

/**
 * Returns the polygon in the given topology with the specified face.
 * @param {module:topo~Topo} topo
 * @param {module:face~Face} face
 * @return {number[][][]}
 */
function getFaceGeometry (topo, face) {
  const edges = getEdgeByFace(topo, face, false);
  if (edges.length === 0) {
    return []
  } else {
    const css = edges.map(e => e.coordinates);
    return polygonize(css)
  }
}

function getInteriorEdgePoint (coordinates) {
  // TODO: too naive
  return coordinates[1]
}

/**
 * @param {module:topo~Topo} topo
 * @param {module:edge~Edge} edge
 * @param {boolean} dir
 * @param {module:face~Face} face
 * @param {boolean} mbrOnly
 * @return {module:face~Face}
 * @private
 */
function addFaceSplit (topo, edge, dir, face, mbrOnly) {
  console.debug(`addFaceSplit called on edge ${sid(edge, dir)} and face ${face.id}`);

  const universe = topo.faces[0];

  const sedges = getRingEdges(topo, edge, dir, 0);

  sedges.forEach((se, i) => console.debug(`Component ${i} in ring of edge ${edge.id} is edge ${sid(se.edge, se.dir)}`));

  if (sedges.length === 0) throw new Error('no ring edges for edge' + edge.id)

  console.debug(`getRingEdges returned ${sedges.length} edges`);

  if (sedges.some(se => se.edge === edge && se.dir === !dir)) {
    console.debug('not a ring');
    return 0
  }

  console.debug(`Edge ${sid(edge, dir)} split face ${face.id} (mbr_only:${mbrOnly})`);

  const newFace = { id: topo.facesSeq };

  // const ringEdges = sedges.map(se => se.edge).filter((elem, pos, arr) => arr.indexOf(elem) === pos)

  sedges.forEach((se, i) => {
    console.debug(`Edge ${i} in ring of edge ${sid(edge, dir)} is edge ${sid(se.edge, se.dir)}`);
  });

  const shell = sedges
    .map(e => e.dir ? e.edge.coordinates : e.edge.coordinates.slice().reverse())
    .reduce((a, b) => a.concat(b), []);

  const isccw = signedArea(shell) <= 0;
  console.debug(`Ring of edge ${edge.id} is ${isccw ? 'counter' : ''}clockwise`);

  if (face === universe) {
    if (!isccw) {
      console.debug('The left face of this clockwise ring is the universe, will not create a new face there.');
      return -1
    }
  }

  if (mbrOnly && face !== universe) {
    return -1
  }

  const newFaceIsOutside = face !== universe && !isccw;

  if (newFaceIsOutside) {
    console.debug('New face is on the outside of the ring, updating rings in former shell');
  } else {
    console.debug('New face is on the inside of the ring, updating forward edges in new ring');
  }

  const faceEdges = getEdgeByFace(topo, face, newFace);

  console.debug(`getEdgeByFace returned ${faceEdges.length} edges`);

  faceEdges.forEach(e => {
    let found = 0;
    sedges.every(se => {
      if (e === se.edge) {
        if (se.dir) {
          console.debug(`Edge ${e.id} is a forward edge of the new ring`);
          e.leftFace = newFace;
        } else {
          console.debug(`Edge ${e.id} is a backward edge of the new ring`);
          e.rightFace = newFace;
        }
        found++;
        if (found === 2) return false
      }
      return true
    });
    if (found > 0) return

    const ep = getInteriorEdgePoint(edge.coordinates);
    const contains = pointInPoly(ep, shell);

    if (newFaceIsOutside) {
      if (contains) {
        console.debug(`Edge ${e.id} contained in an hole of the new face`);
        return
      }
    } else {
      if (!contains) {
        console.debug(`Edge ${e.id} not contained in the face shell`);
        return
      }
    }

    if (e.leftFace === face) {
      console.debug(`Edge ${e.id} has new face on the left side`);
      e.leftFace = newFace;
    }

    if (e.rightFace === face) {
      console.debug(`Edge ${e.id} has new face on the right side`);
      e.rightFace = newFace;
    }
  });

  const nodes = getNodeByFace(topo, face);

  nodes.forEach(n => {
    const contains = pointInPoly(n.coordinate, shell);
    console.debug(`Node ${n.id} is ${contains ? '' : 'not '}contained in new ring, newface is ${newFaceIsOutside ? 'outside' : 'inside'}`);
    if (newFaceIsOutside) {
      if (contains) {
        console.debug(`Node ${n.id} contained in an hole of the new face`);
        return
      }
    } else {
      if (!contains) {
        console.debug(`Node ${n.id} contained in the face shell`);
        return
      }
    }
    n.face = newFace;
  });

  insertFace(topo, newFace);
  updateFaceTree(topo, newFace);

  trigger(topo, 'addface', newFace);

  return newFace
}

/** @module */

/**
 * Node definition
 *
 * @typedef {Object} Node
 * @property {number} id Node ID
 * @property {module:face~Face} face Containing face
 * @property {module:coordinate~Coordinate} coordinate Coordinate
 * @property {number} minX Minimum X of bounds
 * @property {number} maxY Maximum Y of bounds
 * @property {number} minX Minimum X of bounds
 * @property {number} maxY Maximum Y of bounds
 */

/**
 * Find the node at a point location.
 *
 * @param {module:topo~Topo} topo
 * @param {module:coordinate~Coordinate} coordinate
 * @return {module:node~Node}
 */
function getNodeByPoint (topo, coordinate) {
  const result = topo.nodesTree.search({
    minX: coordinate[0],
    minY: coordinate[1],
    maxX: coordinate[0],
    maxY: coordinate[1]
  });
  if (result.length === 0) return
  if (result.length === 1) return result[0]
  throw Error('getNodeByPoint: unexpected search result')
}

/**
 * Adds an isolated node to a face in a topology and returns the new node. If face is null, the node is still created.
 *
 * @param {module:topo~Topo} topo
 * @param {module:coordinate~Coordinate} coordinate
 * @return {module:node~Node}
 */
function addIsoNode (topo, coordinate) {
  const { nodesTree: tree, faces } = topo;

  const containingFaces = getFaceByPoint(topo, coordinate, 0);

  const node = {
    id: topo.nodesSeq,
    face: containingFaces.length === 0 ? faces[0] : containingFaces[0],
    coordinate,
    minX: coordinate[0],
    minY: coordinate[1],
    maxX: coordinate[0],
    maxY: coordinate[1]
  };

  if (!tree.collides(node)) {
    insertNode(topo, node);
    trigger(topo, 'addnode', node);
    return node
  } else {
    throw new SpatialError('coincident node')
  }
}

function removeIsoNode (topo, node) {
  if (!node.face) throw new SpatialError('not isolated node')
  deleteNode(topo, node);
  trigger(topo, 'removenode', node);
}

/** @module */

/**
 * Topology structure
 *
 * @typedef {Object} Topo
 * @property {module:node~Node[]} nodes
 * @property {Object} nodesTree
 * @property {module:edge~Edge[]} edges
 * @property {Object} edgesTree
 * @property {module:face~Face[]} faces
 * @property {Object} facesTree
 * @property {module:face~Face} universe
 * @property {function} getNodeByPoint Calls {@link module:node.getNodeByPoint} in the context of this object
 * @property {function} addIsoNode Calls {@link module:node.addIsoNode} in the context of this object
 * @property {function} removeIsoNode Calls {@link module:node.removeIsoNode} in the context of this object
 * @property {function} getEdgeByPoint Calls {@link module:edge.getEdgeByPoint} in the context of this object
 * @property {function} getEdgesByLine Calls {@link module:edge.getEdgesByLine} in the context of this object
 * @property {function} addIsoEdge Calls {@link module:edge.addIsoEdge} in the context of this object
 * @property {function} addEdgeNewFaces Calls {@link module:edge.addEdgeNewFaces} in the context of this object
 * @property {function} addEdgeModFace Calls {@link module:edge.addEdgeModFace} in the context of this object
 * @property {function} remEdgeNewFace Calls {@link module:edge.remEdgeNewFace} in the context of this object
 * @property {function} remEdgeModFace Calls {@link module:edge.remEdgeModFace} in the context of this object
 * @property {function} newEdgesSplit Calls {@link module:edge.newEdgesSplit} in the context of this object
 * @property {function} modEdgeSplit Calls {@link module:edge.modEdgeSplit} in the context of this object
 * @property {function} newEdgeHeal Calls {@link module:edge.newEdgeHeal} in the context of this object
 * @property {function} getRingEdges Calls {@link module:face.getRingEdges} in the context of this object
 * @property {function} getFaceGeometry Calls {@link module:face.getFaceGeometry} in the context of this object
 * @property {function} on Calls {@link module:topo.on} in the context of this object
 * @property {function} un Calls {@link module:topo.un} in the context of this object
 * @fires module:topo~addface
 * @fires module:topo~removeface
 * @fires module:topo~addedge
 * @fires module:topo~modedge
 * @fires module:topo~removeedge
 * @fires module:topo~addnode
 * @fires module:topo~removenode
 */

/**
 * Emitted when a face has been added to the topology.
 * @event module:topo~addface
 * @type {module:face~Face}
 */

/**
 * Emitted when a face has been removed from the topology.
 * @event module:topo~removeface
 * @type {module:face~Face}
 */

/**
 * Emitted when an edge has been added to the topology.
 * @event module:topo~addedge
 * @type {module:edge~Edge}
 */

/**
 * Emitted when an edge has been modified.
 * @event module:topo~modedge
 * @type {module:edge~Edge}
 */

/**
 * Emitted when an edge has been removed from the topology.
 * @event module:topo~removeedge
 * @type {module:edge~Edge}
 */

/**
 * Emitted when a node has been added to the topology.
 * @event module:topo~addnode
 * @type {module:node~Node}
 */

/**
 * Emitted when a node has been removed from the topology.
 * @event module:topo~removenode
 * @type {module:node~Node}
 */

/**
 * Create topology.
 *
 * @param {string} name
 * @param {number} srid
 * @param {number} tolerance
 * @return {module:topo~Topo}
 */
function createTopology (name, srid, tolerance) {
  const nodes = [];
  const nodesTree = rbush(16);
  const edges = [];
  const edgesTree = rbush(16);
  const universe = { id: 0 };
  const faces = [universe];
  const facesTree = rbush(16);
  const topo = {
    name,
    srid,
    tolerance,
    nodes,
    nodesSeq: 1,
    nodesTree,
    edges,
    edgesSeq: 1,
    edgesTree,
    faces,
    facesSeq: 1,
    facesTree,
    universe,
    getNodeByPoint: (...args) => getNodeByPoint(topo, ...args),
    addIsoNode: (...args) => addIsoNode(topo, ...args),
    removeIsoNode: (...args) => removeIsoNode(topo, ...args),
    getEdgeByPoint: (...args) => getEdgeByPoint(topo, ...args),
    getEdgesByLine: (...args) => getEdgesByLine(topo, ...args),
    addIsoEdge: (...args) => addIsoEdge(topo, ...args),
    addEdgeNewFaces: (...args) => addEdgeNewFaces(topo, ...args),
    addEdgeModFace: (...args) => addEdgeModFace(topo, ...args),
    remEdgeNewFace: (...args) => remEdgeNewFace(topo, ...args),
    remEdgeModFace: (...args) => remEdgeModFace(topo, ...args),
    newEdgesSplit: (...args) => newEdgesSplit(topo, ...args),
    modEdgeSplit: (...args) => modEdgeSplit(topo, ...args),
    newEdgeHeal: (...args) => newEdgeHeal(topo, ...args),
    modEdgeHeal: (...args) => modEdgeHeal(topo, ...args),
    getRingEdges: (...args) => getRingEdges(topo, ...args),
    getFaceGeometry: (...args) => getFaceGeometry(topo, ...args),
    observers: {
      'addface': [],
      'modface': [],
      'removeface': [],
      'addedge': [],
      'modedge': [],
      'removeedge': [],
      'addnode': [],
      'removenode': []
    },
    on: (...args) => on(topo, ...args),
    un: (...args) => un(topo, ...args)
  };
  return topo
}

/**
 * Registers a callback for a named event
 * @param {module:topo~Topo} topo Topology instance.
 * @param {string} name Event name.
 * @param {function} callback Callback function.
 */
function on (topo, name, callback) {
  topo.observers[name].push(callback);
}

/**
 * Unregisters a callback for a named event
 * @param {module:topo~Topo} topo Topology instance.
 * @param {string} name Event name.
 * @param {function} callback Callback function.
 */
function un (topo, name, callback) {
  const i = topo.observers[name].indexOf(callback);
  topo.observers[name].splice(i, 1);
}

function trigger (topo, name, e) {
  topo.observers[name].forEach(o => o(e));
}

function insertFace (topo, face) {
  const { faces } = topo;
  face.id = topo.facesSeq++;
  faces.push(face);
}

function updateFaceTree (topo, face) {
  const { facesTree } = topo;
  const coordinates = getFaceGeometry(topo, face);
  const xs = coordinates[0].map(c => c[0]);
  const ys = coordinates[0].map(c => c[1]);
  face.minX = Math.min(...xs);
  face.minY = Math.min(...ys);
  face.maxX = Math.max(...xs);
  face.maxY = Math.max(...ys);
  facesTree.insert(face);
}

function deleteFace (topo, face) {
  const { faces, facesTree } = topo;
  facesTree.remove(face);
  // delete faces[faces.indexOf(face)]
  faces.splice(faces.indexOf(face), 1);
}

function insertEdge (topo, edge) {
  const { edges, edgesTree } = topo;
  const xs = edge.coordinates.map(c => c[0]);
  const ys = edge.coordinates.map(c => c[1]);
  edge.id = topo.edgesSeq++;
  edge.minX = Math.min(...xs);
  edge.minY = Math.min(...ys);
  edge.maxX = Math.max(...xs);
  edge.maxY = Math.max(...ys);
  edgesTree.insert(edge);
  edges.push(edge);
}

function deleteEdge (topo, edge) {
  const { edges, edgesTree } = topo;
  edgesTree.remove(edge);
  // delete edges[edges.indexOf(edge)]
  edges.splice(edges.indexOf(edge), 1);
}

function insertNode (topo, node) {
  const { nodes, nodesTree } = topo;

  const coordinate = node.coordinate;

  node.id = topo.nodesSeq++;
  node.minX = coordinate[0];
  node.minY = coordinate[1];
  node.maxX = coordinate[0];
  node.maxY = coordinate[1];

  nodesTree.insert(node);
  nodes.push(node);
}

function deleteNode (topo, node) {
  const { nodes, nodesTree } = topo;
  nodesTree.remove(node);
  // delete nodes[nodes.indexOf(node)]
  nodes.splice(nodes.indexOf(node), 1);
}

/**
 * Single file entry point module
 * @module
 */

export { createTopology };
